<?xml version="1.0" encoding="UTF-8"?>
<tileset version="1.8" tiledversion="1.8.2" name="WinColl" tilewidth="16" tileheight="16" tilecount="10" columns="0">
 <grid orientation="orthogonal" width="1" height="1"/>
 <tile id="9">
  <image width="16" height="16" source="32.png"/>
 </tile>
 <tile id="10">
  <image width="16" height="16" source="35.png"/>
 </tile>
 <tile id="11">
  <image width="16" height="16" source="36.png"/>
 </tile>
 <tile id="12">
  <image width="16" height="16" source="42.png"/>
 </tile>
 <tile id="13">
  <image width="16" height="16" source="43.png"/>
 </tile>
 <tile id="14">
  <image width="16" height="16" source="46.png"/>
 </tile>
 <tile id="15">
  <image width="16" height="16" source="64.png"/>
 </tile>
 <tile id="16">
  <image width="16" height="16" source="75.png"/>
 </tile>
 <tile id="17">
  <image width="16" height="16" source="87.png"/>
 </tile>
 <tile id="18">
  <image width="16" height="16" source="32.png"/>
 </tile>
</tileset>
